<?php
/**
* 2007-2015 PrestaShop
*
* NOTICE OF LICENSE
*
* This source file is subject to the Academic Free License (AFL 3.0)
* that is bundled with this package in the file LICENSE.txt.
* It is also available through the world-wide-web at this URL:
* http://opensource.org/licenses/afl-3.0.php
* If you did not receive a copy of the license and are unable to
* obtain it through the world-wide-web, please send an email
* to license@prestashop.com so we can send you a copy immediately.
*
* DISCLAIMER
*
* Do not edit or add to this file if you wish to upgrade PrestaShop to newer
* versions in the future. If you wish to customize PrestaShop for your
* needs please refer to http://www.prestashop.com for more information.
*
*  @author    PrestaShop SA <contact@prestashop.com>
*  @copyright 2007-2015 PrestaShop SA
*  @license   http://opensource.org/licenses/afl-3.0.php  Academic Free License (AFL 3.0)
*  International Registered Trademark & Property of PrestaShop SA
*/

class OfdfermaApi
{
    const API_URL   = 'https://ferma.ofd.ru/api/kkt/cloud';
    const AUTH_URL  = 'https://ferma.ofd.ru/api/Authorization/CreateAuthToken';
    const CHECK_URL = 'https://check.ofd.ru/rec/';
    
    private $debug = false;
    
    private $settings = array();
    
    private static $instance;
    
    public static function getInstance()
    {
        if (null === static::$instance) {
            static::$instance = new static();
        }

        return static::$instance;
    }
    
    function __construct() {

        $this->settings = array(
            'login'         => defined('OFDFERMA_USER') ? OFDFERMA_USER : '',
            'password'      => defined('OFDFERMA_PASS') ? OFDFERMA_PASS : '',
            'osn'           => defined('OFDFERMA_NALOG') ? OFDFERMA_NALOG : '',
            'inn'           => defined('OFDFERMA_INN') ? OFDFERMA_INN : '',
            'nds'           => defined('OFDFERMA_VAT') ? OFDFERMA_VAT : '',
            'collapse'      => defined('OFDFERMA_COLLAPSE') ? OFDFERMA_COLLAPSE : '',
            'collapse_name' => defined('OFDFERMA_COLLAPSE_NAME') ? OFDFERMA_COLLAPSE_NAME : '',
            'status'        => defined('OFDFERMA_STATUS') ? OFDFERMA_STATUS : '',
            'email'         => defined('OFDFERMA_EMAIL') ? OFDFERMA_EMAIL : '',
            
            'token'         => defined('OFDFERMA_TOKEN') ? OFDFERMA_TOKEN : '',
            'token_exp_date'=> defined('OFDFERMA_TOKEN_EXP_DATE') ? OFDFERMA_TOKEN_EXP_DATE : ''
        );
    }
    
    public function setDebug($val) {
        $this->debug = $val;
    }
    
    public function getDataType($type_id = FALSE){
        
        $aResult = array(
            'Income'        => 'Приход',
            'IncomeReturn'  => 'Возврат',
        );
        
        if ($type_id){
            if (isset($aResult[$type_id])){
                $aResult = $aResult[$type_id];
            }else{
                $aResult = $type_id;
            }
        }

        return $aResult;
    }

    //Список НДС
    public function getDataVat() {
        return array(
            'VatNo' => Core::_('Ofdferma.settings_vat_vatno'),
            'Vat0'  => Core::_('Ofdferma.settings_vat_vat0'),
            'Vat10' => Core::_('Ofdferma.settings_vat_vat10'),
            'Vat18' => Core::_('Ofdferma.settings_vat_vat18'),
            'Vat20' => Core::_('Ofdferma.settings_vat_vat20'),
            'CalculatedVat10110' => Core::_('Ofdferma.settings_vat_vat10110'),
            'CalculatedVat18118' => Core::_('Ofdferma.settings_vat_vat18118'),
            'CalculatedVat20120' => Core::_('Ofdferma.settings_vat_vat20120'),
        );
    }

    //Список Систем налогооблажения
    public function getDataNalog() {
        return array(
            'Common' => Core::_('Ofdferma.settings_nalog_common'),
            'SimpleIn' => Core::_('Ofdferma.settings_nalog_simplein'),
            'SimpleInOut' => Core::_('Ofdferma.settings_nalog_simpleinout'),
            'Unified' => Core::_('Ofdferma.settings_nalog_unified'),
            'UnifiedAgricultural' => Core::_('Ofdferma.settings_nalog_unifiedagricultural'),
            'Patent' => Core::_('Ofdferma.settings_nalog_patent'),
        );
    }

    //Список Статусов
    public function getDataStatus() {

        $aData = array();

        $aData[0] = Core::_('Ofdferma.settings_status_no_select');

        $aObjects = Core_Entity::factory('Shop_Order_Status')->findAll();
        foreach ($aObjects as $oObject) {
            $aData[$oObject->id] = $oObject->name;
        }

        $aData['paid'] = Core::_('Ofdferma.settings_status_paid');
        $aData['cancel'] = Core::_('Ofdferma.settings_status_canceled');
        //$aData['new'] = Core::_('Ofdferma.settings_status_new');

        return $aData;
    }
    
    
    private function check_settings() {
        $ofd_login      = $this->settings['login'];
        $ofd_pass       = $this->settings['password'];
        $ofd_osn        = $this->settings['osn'];
        $ofd_inn        = $this->settings['inn'];
        $ofd_nds        = $this->settings['nds'];

        return (
            $ofd_login &&
            $ofd_pass &&
            $ofd_osn &&
            $ofd_inn &&
            $ofd_nds
        );
    }
        
    private function getErrorByCode($code_id, $data) {
        $msg = '';

        SWITCH ($code_id) {
            case 1019:
                $msg = "Идентификатор счета '{$data['Request']['InvoiceId']}' уже существует в ОФД";
                break;
        }

        return $msg;
    }
    
    private function formatFloat( $int, $count = 2){
        return	round( (float)$int , $count);
    }
    
    /**
    * Sanitize phone number
    * @return self
    */
    private function sanitizePhoneNumber($phone)
    {
        $phone = preg_replace('/[^0-9]/', '', $phone);

        // 8928 => 7928
        substr($phone, 0, 2) == '89' && $phone{0} = ' ';

        return trim($phone);
    }
    
    private function customErrorHandler($errno, $errstr, $errfile, $errline, array $errcontext) {
        $this->log_msg($errstr);

        if (0 === error_reporting()) {
            return false;
        }
        throw new Exception($errstr, 0, $errno, $errfile, $errline);
    }
    
    private function get_receipt($order_id, $type){
        
        $oQuery = Core_QueryBuilder::select()
            ->from('ofdferma')
            ->where('order_id', '=', (int)$order_id)
            ->where('type', '=', addslashes($type))
            ->limit(1);

        $aResults = $oQuery->execute()->asAssoc()->result();
        
        return (bool)$aResults;
    }
    
    //Обновляем токен
    private function get_token(){
        
        if ( ! $this->check_settings()) {
            return $this->log_msg('Проверьте настройки, для корректной работы должны быть заполнены: Логин, Пароль, НДС по умолчанию, ИНН, Система налогообложения');
        };
        
        if ($this->settings['token'] && ($this->settings['token_exp_date'] > (time() - 10))) {
            return $this->log_msg('', true);
        }
        
        $ofd_login      = $this->settings['login'];
        $ofd_pass       = $this->settings['password'];
        
        $ofd_auth_url   = self::AUTH_URL;
        
        $response = $this -> execute_request($ofd_auth_url, array(
            "Login"     => $ofd_login,
            "Password"  => $ofd_pass,
        ));
        
        if (isset($response['result'])){
            $result = $response['result'];
            
            //Обновляем токен
            $this->settings['token'] = $result['Data']['AuthToken'];
            $this->settings['token_exp_date'] = $result['Data']['ExpirationDateUtc'];
            
            //Token
            $oConstantTmp = Core_Entity::factory('Constant')->getByName('OFDFERMA_TOKEN');
            if (is_null($oConstantTmp))
            {
                $oConstantTmp = Core_Entity::factory('Constant');
                $oConstantTmp->name = 'OFDFERMA_TOKEN';
                $oConstantTmp->active = 1;
            }
            $oConstantTmp->value = $this->settings['token'];
            $oConstantTmp->save();

            //Exp Date
            $oConstantTmp = Core_Entity::factory('Constant')->getByName('OFDFERMA_TOKEN_EXP_DATE');
            if (is_null($oConstantTmp))
            {
                $oConstantTmp = Core_Entity::factory('Constant');
                $oConstantTmp->name = 'OFDFERMA_TOKEN_EXP_DATE';
                $oConstantTmp->active = 1;
            }
            $oConstantTmp->value = $this->settings['token_exp_date'];
            $oConstantTmp->save();
            
            return $this->log_msg('', true);
        }
        
        return $response;
    }
    
    private function execute_request($url, $data = null) {  
        
        $options = array(
            "ssl" => array(
                "verify_peer" => false,
                "verify_peer_name" => false,
            ),
            'http' => array(
                'timeout' => 10,
                'ignore_errors' => true,
                'content' => json_encode($data),
                'header' => "Content-type: application/json\r\n" .
                "Accept: application/json" . "\r\n",
                "Content-Length: " . strlen(json_encode($data)) . "\r\n",
                'method' => 'POST',
            )
        );
        
        $context = stream_context_create($options);
        set_error_handler(array($this, 'customErrorHandler'));

        try {
            $result = file_get_contents($url, false, $context);
        } catch (Exception $e) {
            return $this->log_msg($e->getMessage());
        }
        
        restore_error_handler();
        $result = json_decode($result, TRUE);
        
        if (isset($result['Status']) && ($result['Status'] == 'Success')) {
            return $this->log_msg('', true, array('result' => $result));
        } else if (isset($result['Status']) && ($result['Status'] == 'Failed')) {
            $msg = $result['Error']['Message'];
            if ($result['Error']['Code']) {
                if ($val = $this->getErrorByCode($result['Error']['Code'], $data)) {
                    $msg = $val;
                }
            }
            
            return $this->log_msg($msg, false, $result);
        } else {
            return $this->log_msg('some error', false, $result);
        }
    }
    
    
    public function getCheckLink($row){
        $check_link = self::CHECK_URL . "{$row->inn}/{$row->rnm}/{$row->fn}/".ltrim($row->fdn,0)."/".$row->fpd;
        
        return $check_link;
    }
    
    public function get_product_vat($product, $default = ''){
        
        $aResults = Core_DataBase::instance()->query('SELECT * FROM ofdferma_link_product WHERE item_id = "' .(int)$product->shop_item_id. '" LIMIT 1')->asObject(null)->result(FALSE);
        $nds = isset($aResults[0]->vat) ? $aResults[0]->vat : '';
        
        return $nds ? $nds : $default;
    }
    
    public function log_msg($mess, $status = false, $params = array()) {
        if ( ! $status && $this->debug && $mess) {
            $txt = $mess;
            
            Core_Log::instance()->clear()
                ->status(Core_Log::$ERROR)
                ->write($txt);
            
            $user_mail = $this->settings['email'];
            if ($user_mail){                  
                Core_Mail::instance()
                    ->clear()
                    ->to($user_mail)
                    ->from(defined('EMAIL_TO') ? EMAIL_TO : SUPERUSER_EMAIL)
                    ->subject('OFD Ferma - Ошибка')
                    ->message(date('d.m.Y H:i:s') .' : '. $txt)
                    ->contentType('text/plain')
                    ->send();
            }
        }
        
        return array_merge(array(
            'status'    => $status,
            'mess'      => 'FERMA OFD.ru: ' . $mess
        ), (array)$params);
    } 
    
    public function set_status_order($order_id, $state_ids){
        
        $res = FALSE;
        //Если статус является тем при котором нужно формировать чек прихода
        if ($state_ids && in_array($this->settings['status'], $state_ids)){
            $res = array('status' => false);
            
            $this->setDebug(1);
            $res = $this->Check($order_id, 'Income');
            $this->setDebug(0);
        }
        
        return $res;
    }
    
    /**
    * Отправка чека в ОФД Ferma
    * @param array $data данные заказа
    * @param string $operation операция [sale|sale_refund]
    * @param bool $json передача упакованных внешних данных в JSON формате
    */
    public function Check($order_id, $type = 'Income') {
       
        $res = $this->get_token();
        if ( ! $res['status']) {
            return $res;
        }
       
        $oOrder = Core_Entity::factory('Shop_order', $order_id);
        $order_items = $oOrder->Shop_Order_Items->findAll(FALSE);

        $temptype = $this -> getDataType($type);
        
        // Данные из заказа
        if ( ! $order_items) {
            return $this->log_msg("В заказе #{$order_id} нет товаров");
        }
        
        if ($this -> get_receipt($order_id, $type)){
            return $this->log_msg("Чек {$temptype} для заказа {$order_id} уже существует в модуле");
        }
        
        //Формируем данные для запроса
        $prefix     = 'o';
        
        $ofd_osn    = $this->settings['osn'];
        $ofd_inn    = $this->settings['inn'];
        $ofd_nds    = $this->settings['nds'];
        
        //свертка
        $ofd_collapse       = $this->settings['collapse'];
        $ofd_collapse_name  = $this->settings['collapse_name'];
        
        $customer_phone = str_replace(array('(', ')', '-', ' '), '', $this -> sanitizePhoneNumber($oOrder->phone));
        $customer_email = $oOrder->email;     

        //Формируем данные
        $data = array();
        $data['Request']['Inn'] = $ofd_inn;
        $data['Request']['Type'] = $type;
        $data['Request']['InvoiceId'] = $prefix . $oOrder->guid . '-' . $type;
        $data['Request']['LocalDate'] = date('Y-m-d\TH:i:s');
        $data['Request']['CustomerReceipt'] = array(
            'TaxationSystem'    => $ofd_osn,
            'Email'             => $customer_email,
            'Phone'             => $customer_phone,
            'Items'             => array(),
        );
           
        
        //Товары
        if (is_array($order_items)) {
            foreach ($order_items as $shop_order_item){   
                $vat = $this->get_product_vat($shop_order_item, $ofd_nds);
                
                array_push($data['Request']['CustomerReceipt']['Items'], array(
                        'Label'     => $shop_order_item->name,
                        'Price'     => $this->formatFloat( $shop_order_item->price ),
                        'Quantity'  => $this->formatFloat( $shop_order_item->quantity, 3),
                        'Amount'    => $this->formatFloat( $shop_order_item->price * $shop_order_item->quantity ),
                        'Vat'       => $vat,
                    )
                );
            }
        }        
        
        //Если включена свертка
        if ($ofd_collapse) {
            $pos_name = $ofd_collapse_name ? $ofd_collapse_name : 'Undefined';

            $sum = 0;
            foreach ($data['Request']['CustomerReceipt']['Items'] as $item) {
                $sum += $item['Amount'];
            }
            $data['Request']['CustomerReceipt']['Items'] = array(array(
                'Label'     => $pos_name,
                'Price'     => $sum,
                'Quantity'  => 1,
                'Amount'    => $sum,
                'Vat'       => $ofd_nds,
            ));
        }
        
        
        //Отправляем запрос к ОФД
        $api_url = self::API_URL . "/receipt?AuthToken=" . $this->settings['token'];
        $response = $this -> execute_request($api_url, $data);
       
        if (isset($response['result'])){
            $result = $response['result'];
            
            $aInsertData = array(
                'site_id'                   => defined('CURRENT_SITE') ? CURRENT_SITE : 0,
                'shop_id'                   => $oOrder->shop_id,
                'num'                       => $result['Data']['ReceiptId'],
                'type'                      => $data['Request']['Type'],
                'inn'                       => $data['Request']['Inn'],
                'order_id'                  => $order_id,
                'total'                     => $oOrder->getAmount(),
                'created_at'                => gmdate('Y-m-d H:i:s')
            );
            
            // Запись лога
            Core_DataBase::instance()
                ->query('INSERT INTO ofdferma ('.implode(',', array_map(array(Core_DataBase::instance(), 'quoteColumnName'),array_keys($aInsertData))).') '
                      . 'VALUES('.implode(',', array_map(array(Core_DataBase::instance(), 'quote'),array_values($aInsertData))).')');
                 
            return $this->log_msg("Чек {$temptype} для заказа {$order_id} сформирован: {$result['Data']['ReceiptId']}", true);
        }else{
            return $response;
        }
    }
   
    //Обновляем статус чеков
    public function check_update($id = FALSE){
        
        $res = $this->get_token();
        if ( ! $res['status']) {
            return $res;
        }

        $sWhere = '';
        if ($id){
            $sWhere = ' && `num` = "' .addslashes($id). '"';
        }
	
        if ($results = Core_DataBase::instance()->query('SELECT * FROM ofdferma WHERE `status` NOT IN ("FAILED","CONFIRMED")'.$sWhere)->asObject(null)->result(FALSE)){
			
            $this -> setDebug(1);
            foreach ($results as $result) {
                    $data = array();
                    $data['Request']['ReceiptId'] = $result->num;

                    //Отправляем запрос к ОФД
                    $api_url = self::API_URL . "/status?AuthToken=" . $this->settings['token'];
                    $response = $this -> execute_request($api_url, $data);

                    if (isset($response['result'])){
                        $aDataUpdateTmp = array(
                            'status'                    => $response['result']['Data']['StatusName'],
                            'status_message'            => $response['result']['Data']['StatusMessage'],
                            'fn'                        => $response['result']['Data']['Device']['FN'],
                            'rnm'                       => $response['result']['Data']['Device']['RNM'],
                            'fdn'                       => $response['result']['Data']['Device']['FDN'],
                            'fpd'                       => $response['result']['Data']['Device']['FPD'],
                            'updated_at'                => gmdate('Y-m-d H:i:s')
                        );
                        
                        $aDataUpdate = array();
                        foreach($aDataUpdateTmp as $key => $val){
                            $aDataUpdate[] = "`{$key}` = '".addslashes($val)."'";
                        }
                        
                        // Запись лога    
                        Core_DataBase::instance()
                            ->query('UPDATE ofdferma SET '
                                  . implode(',', $aDataUpdate)
                                  . 'WHERE `num` = "' .addslashes($result->num). '"');
                    }
            }
            $this -> setDebug(0);
        }
    }
}
